From: Roy Funderburk <royffmpeg@funderburk.us>
Date: Sat, 15 Apr 2023 13:04:42 -0700
Subject: [PATCH] avformat: Add DTS-UHD demuxer and movenc support.
X-Unsent: 1
To: ffmpeg-devel@ffmpeg.org

Demuxing of DTS-UHD input files per ETSI TS 102 114 is added as
demuxer "dtsuhd".  movenc supports DTS-UHD audio track.

Signed-off-by: Roy Funderburk <Roy.Funderburk@xperi.com>
---
 Changelog                 |   1 +
 configure                 |   1 +
 doc/general_contents.texi |   1 +
 libavformat/Makefile      |   1 +
 libavformat/allformats.c  |   1 +
 libavformat/dtshddec.c    |  25 +++++-
 libavformat/dtsuhddec.c   | 163 ++++++++++++++++++++++++++++++++++++++
 libavformat/movenc.c      |  32 ++++++++
 8 files changed, 222 insertions(+), 3 deletions(-)
 create mode 100644 libavformat/dtsuhddec.c

diff --git a/Changelog b/Changelog
index cd872d9bf2..31a11385df 100644
--- a/Changelog
+++ b/Changelog
@@ -17,6 +17,7 @@ version <next>:
 - RivaTuner video decoder
 - xfade_vulkan filter
 - vMix video decoder
+- DTS-UHD demuxer
 
 version 6.0:
 - Radiance HDR image support
diff --git a/configure b/configure
index 4ac7cc6c0b..bcd770f6f0 100755
--- a/configure
+++ b/configure
@@ -3484,6 +3484,7 @@ dash_demuxer_deps="libxml2"
 dirac_demuxer_select="dirac_parser"
 dts_demuxer_select="dca_parser"
 dtshd_demuxer_select="dca_parser"
+dtsuhd_demuxer_select="dtsuhd_parser"
 dv_demuxer_select="dvprofile"
 dv_muxer_select="dvprofile"
 dxa_demuxer_select="riffdec"
diff --git a/doc/general_contents.texi b/doc/general_contents.texi
index 2eeebd847d..e1ba9c4597 100644
--- a/doc/general_contents.texi
+++ b/doc/general_contents.texi
@@ -597,6 +597,7 @@ library:
 @item raw DNxHD                 @tab X @tab X
 @item raw DTS                   @tab X @tab X
 @item raw DTS-HD                @tab   @tab X
+@item raw DTS-UHD               @tab   @tab
 @item raw E-AC-3                @tab X @tab X
 @item raw FLAC                  @tab X @tab X
 @item raw GSM                   @tab   @tab X
diff --git a/libavformat/Makefile b/libavformat/Makefile
index 05434a0f82..8839a9d0cc 100644
--- a/libavformat/Makefile
+++ b/libavformat/Makefile
@@ -186,6 +186,7 @@ OBJS-$(CONFIG_DSICIN_DEMUXER)            += dsicin.o
 OBJS-$(CONFIG_DSS_DEMUXER)               += dss.o
 OBJS-$(CONFIG_DTSHD_DEMUXER)             += dtshddec.o
 OBJS-$(CONFIG_DTS_DEMUXER)               += dtsdec.o rawdec.o
+OBJS-$(CONFIG_DTSUHD_DEMUXER)            += dtsuhddec.o
 OBJS-$(CONFIG_DTS_MUXER)                 += rawenc.o
 OBJS-$(CONFIG_DV_MUXER)                  += dvenc.o
 OBJS-$(CONFIG_DVBSUB_DEMUXER)            += dvbsub.o rawdec.o
diff --git a/libavformat/allformats.c b/libavformat/allformats.c
index 96443a7272..89b97653c7 100644
--- a/libavformat/allformats.c
+++ b/libavformat/allformats.c
@@ -144,6 +144,7 @@ extern const AVInputFormat  ff_dss_demuxer;
 extern const AVInputFormat  ff_dts_demuxer;
 extern const FFOutputFormat ff_dts_muxer;
 extern const AVInputFormat  ff_dtshd_demuxer;
+extern const AVInputFormat  ff_dtsuhd_demuxer;
 extern const AVInputFormat  ff_dv_demuxer;
 extern const FFOutputFormat ff_dv_muxer;
 extern const AVInputFormat  ff_dvbsub_demuxer;
diff --git a/libavformat/dtshddec.c b/libavformat/dtshddec.c
index a3dea0668f..8e0cd93b79 100644
--- a/libavformat/dtshddec.c
+++ b/libavformat/dtshddec.c
@@ -39,15 +39,34 @@
 #define STRMDATA 0x5354524D44415441
 #define TIMECODE 0x54494D45434F4445
 
+#define DTSHD_CHUNKHDR_SIZE 16
+#define DTSUHD_SYNC 0x40411BF2
+
 typedef struct DTSHDDemuxContext {
     uint64_t    data_end;
 } DTSHDDemuxContext;
 
 static int dtshd_probe(const AVProbeData *p)
 {
-    if (AV_RB64(p->buf) == DTSHDHDR)
-        return AVPROBE_SCORE_MAX;
-    return 0;
+    int offset = 0;
+    int score = 0;
+    uint64_t chunk_type, chunk_size;
+
+    if (p->buf_size >= DTSHD_CHUNKHDR_SIZE && AV_RB64(p->buf) == DTSHDHDR) {
+        while (offset + DTSHD_CHUNKHDR_SIZE + 4 < p->buf_size) {
+            chunk_type = AV_RB64(p->buf + offset);
+            chunk_size = AV_RB64(p->buf + offset + 8);
+            if (chunk_type == STRMDATA) {
+                // DTSUHD (.dtsx) content is also stored in DTSHDHDR files
+                if (AV_RB32(p->buf + offset + DTSHD_CHUNKHDR_SIZE) != DTSUHD_SYNC)
+                    score = AVPROBE_SCORE_MAX;
+                break;
+            }
+            offset += chunk_size + DTSHD_CHUNKHDR_SIZE;
+        }
+    }
+
+    return score;
 }
 
 static int dtshd_read_header(AVFormatContext *s)
diff --git a/libavformat/dtsuhddec.c b/libavformat/dtsuhddec.c
new file mode 100644
index 0000000000..dfe0236e0f
--- /dev/null
+++ b/libavformat/dtsuhddec.c
@@ -0,0 +1,163 @@
+/*
+ * DTS-UHD audio demuxer
+ * Copyright (c) 2023 Xperi Corporation / DTS, Inc.
+ *
+ * This file is part of FFmpeg.
+ *
+ * FFmpeg is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * FFmpeg is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with FFmpeg; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+ */
+
+/**
+ * @file
+ * Report DTS-UHD audio stream configuration and extract raw packet data.
+ */
+
+#include "internal.h"
+#include "libavcodec/dtsuhd_common.h"
+#include "libavcodec/put_bits.h"
+#include "libavutil/intreadwrite.h"
+
+#define DTSUHD_BUFFER_SIZE (1024 * 1024)
+
+typedef struct DTSUHDDemuxContext {
+    size_t data_end;
+    struct DTSUHD *dtsuhd;
+    uint8_t *buf;
+} DTSUHDDemuxContext;
+
+static int probe(const AVProbeData *p)
+{
+    int offset = av_dtsuhd_strmdata_payload(p->buf, p->buf_size, NULL);
+
+    if (offset >= 0 && offset + 4 < p->buf_size && AV_RB32(p->buf + offset) == DTSUHD_SYNCWORD)
+        return AVPROBE_SCORE_MAX - 3;
+
+    return 0;
+}
+
+static av_cold int read_close(AVFormatContext *s)
+{
+    DTSUHDDemuxContext *dtsxs = s->priv_data;
+
+    av_freep(&dtsxs->buf);
+    av_dtsuhd_destroy(dtsxs->dtsuhd);
+    dtsxs->dtsuhd = NULL;
+
+    return 0;
+}
+
+static int find_first_syncword(DTSUHDDemuxContext *dtsuhd, int data_start)
+{
+    while (data_start + 4 < DTSUHD_BUFFER_SIZE &&
+        !dtsuhd_is_syncword(AV_RB32(dtsuhd->buf + data_start))) {
+        data_start++;
+    }
+
+    return data_start;
+}
+
+static int read_header(AVFormatContext *s)
+{
+    AVIOContext *pb = s->pb;
+    AVStream *st = avformat_new_stream(s, NULL);
+    DTSUHDDemuxContext *dtsuhd = s->priv_data;
+    int buf_bytes;
+    int ret = DTSUHD_INVALID_FRAME;
+    int data_start;
+    int udts_size = 0;
+    uint8_t *udts = NULL;
+
+    if (!(pb->seekable & AVIO_SEEKABLE_NORMAL))
+        return AVERROR(EIO);
+
+    dtsuhd->buf = av_malloc(DTSUHD_BUFFER_SIZE);
+    dtsuhd->dtsuhd = av_dtsuhd_create();
+    if (!dtsuhd->buf || !dtsuhd->dtsuhd || !st)
+        return AVERROR(ENOMEM);
+
+    buf_bytes = avio_read(pb, dtsuhd->buf, DTSUHD_BUFFER_SIZE);
+    if (buf_bytes < 0)
+        return buf_bytes;
+
+    data_start = av_dtsuhd_strmdata_payload(dtsuhd->buf, buf_bytes, &dtsuhd->data_end);
+    if (data_start < 0)
+        return data_start;
+
+    dtsuhd->data_end += data_start;
+    if (data_start == 0)
+        dtsuhd->data_end = avio_size(pb); // Not a DTSHDHDR chunk file, decode frames to end of file.
+
+    data_start = find_first_syncword(dtsuhd, data_start);
+    if (avio_seek(pb, data_start, SEEK_SET) < 0)
+        return AVERROR(EINVAL);
+
+    ret = av_dtsuhd_frame(dtsuhd->dtsuhd, dtsuhd->buf + data_start,
+        buf_bytes - data_start, st->codecpar, &udts, &udts_size);
+    if (ret != DTSUHD_OK) {
+        av_log(s, AV_LOG_ERROR, "Unable to process DTS-UHD file. File may be invalid.\n");
+        return AVERROR_INVALIDDATA;
+    }
+
+    ffstream(st)->need_parsing = AVSTREAM_PARSE_FULL_RAW;
+    st->codecpar->codec_id     = s->iformat->raw_codec_id;
+
+    // Store 'udts' MP4 box for movenc.c
+    if (udts) {
+        ret = ff_alloc_extradata(st->codecpar, udts_size);
+        if (ret == 0)
+            memcpy(st->codecpar->extradata, udts, udts_size);
+        av_freep(&udts);
+        if (ret < 0)
+            return ret;
+    }
+
+    if (st->codecpar->sample_rate)
+        avpriv_set_pts_info(st, 64, 1, st->codecpar->sample_rate);
+
+    return 0;
+}
+
+static int read_packet(AVFormatContext *s, AVPacket *pkt)
+{
+    DTSUHDDemuxContext *dtsuhd = s->priv_data;
+    int64_t size, left;
+    int ret;
+
+    left = dtsuhd->data_end - avio_tell(s->pb);
+    size = FFMIN(left, DTSUHD_MAX_FRAME_SIZE);
+    if (size <= 0)
+        return AVERROR_EOF;
+
+    ret = av_get_packet(s->pb, pkt, size);
+    if (ret < 0)
+        return ret;
+
+    pkt->stream_index = 0;
+
+    return ret;
+}
+
+AVInputFormat ff_dtsuhd_demuxer = {
+    .name           = "dtsuhd",
+    .long_name      = NULL_IF_CONFIG_SMALL("DTS-UHD"),
+    .priv_data_size = sizeof(DTSUHDDemuxContext),
+    .read_probe     = probe,
+    .read_header    = read_header,
+    .read_packet    = read_packet,
+    .read_close     = read_close,
+    .flags          = AVFMT_GENERIC_INDEX,
+    .extensions     = "dtsx",
+    .raw_codec_id   = AV_CODEC_ID_DTSUHD,
+};
diff --git a/libavformat/movenc.c b/libavformat/movenc.c
index a9c8e104f0..2075b3db57 100644
--- a/libavformat/movenc.c
+++ b/libavformat/movenc.c
@@ -762,6 +762,24 @@ static int mov_write_esds_tag(AVIOContext *pb, MOVTrack *track) // Basic
     return update_size(pb, pos);
 }
 
+static int mov_write_udts_tag(AVIOContext *pb, MOVTrack *track)
+{
+    if (track->vos_len < 12) {
+        av_log(pb, AV_LOG_ERROR,
+               "Cannot write moov atom before DTS-UHD packets."
+               " Set the delay_moov flag to fix this.\n");
+        return AVERROR(EINVAL);
+    }
+
+    /* Write vos_data is udts box. */
+    if (memcmp(track->vos_data + 4, "udts", 4) == 0) {
+        avio_write(pb, track->vos_data, track->vos_len);
+        return track->vos_len;
+    }
+
+    return 0;
+}
+
 static int mov_pcm_le_gt16(enum AVCodecID codec_id)
 {
     return codec_id == AV_CODEC_ID_PCM_S24LE ||
@@ -1367,6 +1385,8 @@ static int mov_write_audio_tag(AVFormatContext *s, AVIOContext *pb, MOVMuxContex
         ret = mov_write_dops_tag(s, pb, track);
     else if (track->par->codec_id == AV_CODEC_ID_TRUEHD)
         ret = mov_write_dmlp_tag(s, pb, track);
+    else if (track->par->codec_id == AV_CODEC_ID_DTSUHD)
+        ret = mov_write_udts_tag(pb, track);
     else if (tag == MOV_MP4_IPCM_TAG || tag == MOV_MP4_FPCM_TAG) {
         if (track->par->ch_layout.nb_channels > 1)
             ret = mov_write_chnl_tag(s, pb, track);
@@ -2781,6 +2801,7 @@ static int mov_write_stbl_tag(AVFormatContext *s, AVIOContext *pb, MOVMuxContext
     if ((track->par->codec_type == AVMEDIA_TYPE_VIDEO ||
          track->par->codec_id == AV_CODEC_ID_TRUEHD ||
          track->par->codec_id == AV_CODEC_ID_MPEGH_3D_AUDIO ||
+         track->par->codec_id == AV_CODEC_ID_DTSUHD ||
          track->par->codec_tag == MKTAG('r','t','p',' ')) &&
         track->has_keyframes && track->has_keyframes < track->entry)
         mov_write_stss_tag(pb, track, MOV_SYNC_SAMPLE);
@@ -5676,6 +5697,14 @@ static void mov_parse_vc1_frame(AVPacket *pkt, MOVTrack *trk)
     }
 }
 
+static void mov_parse_dtsuhd_frame(AVPacket *pkt, MOVTrack *trk)
+{
+    if (pkt->size > 4 && AV_RB32(pkt->data) == 0x40411BF2) {
+        trk->cluster[trk->entry].flags |= MOV_SYNC_SAMPLE;
+        trk->has_keyframes++;
+    }
+ }
+
 static void mov_parse_truehd_frame(AVPacket *pkt, MOVTrack *trk)
 {
     int length;
@@ -6346,6 +6375,8 @@ int ff_mov_write_packet(AVFormatContext *s, AVPacket *pkt)
         mov_parse_vc1_frame(pkt, trk);
     } else if (par->codec_id == AV_CODEC_ID_TRUEHD) {
         mov_parse_truehd_frame(pkt, trk);
+    } else if (par->codec_id == AV_CODEC_ID_DTSUHD) {
+        mov_parse_dtsuhd_frame(pkt, trk);
     } else if (pkt->flags & AV_PKT_FLAG_KEY) {
         if (mov->mode == MODE_MOV && par->codec_id == AV_CODEC_ID_MPEG2VIDEO &&
             trk->entry > 0) { // force sync sample for the first key frame
@@ -7803,6 +7834,7 @@ static const AVCodecTag codec_mp4_tags[] = {
     { AV_CODEC_ID_AC3,             MKTAG('a', 'c', '-', '3') },
     { AV_CODEC_ID_EAC3,            MKTAG('e', 'c', '-', '3') },
     { AV_CODEC_ID_DTS,             MKTAG('m', 'p', '4', 'a') },
+    { AV_CODEC_ID_DTSUHD,          MKTAG('d', 't', 's', 'x') },
     { AV_CODEC_ID_TRUEHD,          MKTAG('m', 'l', 'p', 'a') },
     { AV_CODEC_ID_FLAC,            MKTAG('f', 'L', 'a', 'C') },
     { AV_CODEC_ID_OPUS,            MKTAG('O', 'p', 'u', 's') },
-- 
2.17.1

